/***********************************************************

 Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

 SearchLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 SearchLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

 *************************************************************/

/*
 * ChainManager.cpp
 *
 *  Created on: Jun 22, 2011
 *      Author: derrickstolee
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "ChainManager.hpp"
#include "chainfunctions.hpp"

int verbose = 0;

ChainManager::ChainManager(int k, int j, int minN, int maxN, int a, int maxSum, int interval)
{
	this->k = k;
	this->j = j;
	this->interval = interval;
	this->numIntervals = (maxN - minN + interval - 1) / interval;
	this->update_interval = 1000;
	this->start_time = time(NULL);

	switch ( k )
	{
	case 2:
		this->chainFunct = &chainK2N000;
		break;

	case 3:
		this->chainFunct = &chainK3N000;
		break;

	case 4:
		switch ( j )
		{
		case 0:
			this->chainFunct = &chainK4N000;
			break;

		case 1:
			this->chainFunct = &chainK4N001;
			break;

		case 2:
			this->chainFunct = &chainK4N002;
			break;
		}
		break;

	case 5:
		switch ( j )
		{
		case 0:
			this->chainFunct = &chainK5N000;
			break;

		case 1:
			this->chainFunct = &chainK5N001;
			break;

		case 2:
			this->chainFunct = &chainK5N002;
			break;

		case 3:
			this->chainFunct = &chainK5N003;
			break;

		case 4:
			this->chainFunct = &chainK5N004;
			break;

		default:
			printf("--OUT OF RANGE FOR K = %d... 0-4\n", k);
			exit(1);
		}
		break;

	case 6:
		switch ( j )
		{
		case 0:
			this->chainFunct = &chainK6N000;
			break;

		case 1:
			this->chainFunct = &chainK6N001;
			break;

		case 2:
			this->chainFunct = &chainK6N002;
			break;

		case 3:
			this->chainFunct = &chainK6N003;
			break;

		case 4:
			this->chainFunct = &chainK6N004;
			break;

		case 5:
			this->chainFunct = &chainK6N005;
			break;

		case 6:
			this->chainFunct = &chainK6N006;
			break;

		case 7:
			this->chainFunct = &chainK6N007;
			break;

		case 8:
			this->chainFunct = &chainK6N008;
			break;

		case 9:
			this->chainFunct = &chainK6N009;
			break;

		case 10:
			this->chainFunct = &chainK6N010;
			break;

		case 11:
			this->chainFunct = &chainK6N011;
			break;

		case 12:
			this->chainFunct = &chainK6N012;
			break;

		case 13:
			this->chainFunct = &chainK6N013;
			break;

		case 14:
			this->chainFunct = &chainK6N014;
			break;

		case 15:
			this->chainFunct = &chainK6N015;
			break;

		case 16:
			this->chainFunct = &chainK6N016;
			break;

		case 17:
			this->chainFunct = &chainK6N017;
			break;

		default:
			printf("--OUT OF RANGE FOR K = %d... 0-17\n", k);
			exit(1);
		}
		break;

	}

	if ( (minN % 2) == 1 )
	{
		this->minN = minN - 1;
	}
	else
	{
		this->minN = minN;
	}

	this->maxN = maxN;
	this->a = a;
	this->maxSum = maxSum;

	this->num_parameters = 2 * k + 2;
	this->parameters = (int*) malloc(this->num_parameters * sizeof(int));

	for ( int i = 0; i < this->num_parameters; i++ )
	{
		this->parameters[i] = 0;
	}

	this->param_sum = 0;

	int found_size = (this->maxN - this->minN + 6) / 6;
	this->num_missing = (this->maxN - this->minN) / 2;
	this->found_numbers = (char*) malloc(found_size);

	bzero(this->found_numbers, found_size);

	this->cur_pos = -1;
}

ChainManager::~ChainManager()
{
	free(this->parameters);
	free(this->found_numbers);
}

/**
 * pushNext -- deepen the search to the next child
 * 	of the current node.
 *
 * @return the label for the new node. -1 if none.
 */
LONG_T ChainManager::pushNext()
{
	if ( this->cur_pos >= this->num_parameters )
	{
		return -1;
	}

	LONG_T child = 0;
	if ( this->stack.size() > 0 )
	{
		SearchNode* parent = this->stack.back();
		child = parent->curChild + 1;
	}
	else
	{
		child = this->root->curChild + 1;
	}

	if ( child < this->parameters[this->cur_pos] )
	{
		child = this->parameters[this->cur_pos];
	}

	return this->pushTo(child);
}

/**
 * pushTo -- deepen the search to the specified child
 * 	of the current node.
 *
 * @param child the specified label for the new node
 * @return the label for the new node. -1 if none, or failed.
 */
LONG_T ChainManager::pushTo(LONG_T child)
{
	if ( this->cur_pos >= this->num_parameters )
	{
		return -1;
	}

	if ( this->num_missing <= 0)
	{
		return -1;
	}

	if ( this->start_time <= 0 )
	{
		this->start_time = time(NULL);
		this->update_counter = 0;
	}

	if ( this->update_counter > 1000 )
	{
		this->update_counter = 0;
		if ( time(NULL) > this->start_time + this->update_interval )
		{
			this->start_time = time(NULL);
			this->writeMissingList(10);
		}
	}
	else
	{
		this->update_counter = this->update_counter + 1;
	}

	if ( child < 0 || child > this->a )
	{
		return -1;
	}

	//	for ( int i = 0; i < cur_pos; i++ )
	//		printf(" ");
	//	printf("%d\n", (int) child);

	if ( this->cur_pos < 0 )
	{
		/* pick the interval! */
		this->curInterval = child;
		this->curMin = child * this->interval;
		this->curMax = (child + 1) * this->interval;

		if ( this->curMin > this->maxN )
		{
			/* oops! can't augment here */
			return 0;
		}

		this->cur_pos = 0;

		return child;
	}

	this->param_sum = this->param_sum - this->parameters[this->cur_pos];
	this->parameters[this->cur_pos] = (int) child;
	this->param_sum = this->param_sum + this->parameters[this->cur_pos];

	this->cur_pos = this->cur_pos + 1;

	if ( this->stack.size() > 0 )
	{
		SearchNode* parent = this->stack.back();
		parent->curChild = child;
	}
	else
	{
		this->root->curChild = child;
	}

	this->stack.push_back(new SearchNode(child));

	return child;
}

/**
 * pop -- remove the current node and move up the tree.
 *
 * @return the label of the node after the pop.
 * 		This return value is used for validation purposes
 * 		to check proper implementation of push*() and pop().
 */
LONG_T ChainManager::pop()
{
	this->cur_pos = this->cur_pos - 1;

	if ( this->stack.size() > 0 )
	{
		SearchNode* node = this->stack.back();
		this->stack.pop_back();

		delete node;
	}

	if ( this->cur_pos >= 0 )
	{
		int val = this->parameters[this->cur_pos];

		this->param_sum = this->param_sum - val;

		this->parameters[this->cur_pos] = 0;

		return val;
	}
	else
	{
		return this->curInterval;
	}
}

/**
 * prune -- Perform a check to see if this node should be pruned.
 *
 * @return 0 if no prune should occur, 1 if prune should occur.
 */
int ChainManager::prune()
{
	if ( this->cur_pos > this->num_parameters )
	{
		return 1;
	}
	else if ( this->param_sum > this->maxSum )
	{
		return 1;
	}
	else if ( this->chainFunct(this->parameters) > this->maxN )
	{
		return 1;
	}

	return 0;
}

/**
 * isSolution -- Perform a check to see if a solution exists
 * 		at this point.
 *
 * @return 0 if no solution is found, 1 if a solution is found.
 */
int ChainManager::isSolution()
{
	if ( this->cur_pos >= this->num_parameters )
	{
		int num_chains = this->chainFunct(this->parameters);

		this->setNumberFound(num_chains);

		if ( this->num_missing <= 0 )
		{
			this->writeMissingList(1000);
			return -1;
		}
	}

	return 0;
}

/**
 * writeSolution -- create a buffer that contains a
 * 	description of the solution.
 *
 * Solution strings start with S.
 *
 * @return a string of data allocated with malloc().
 * 	It will be deleted using free().
 */
char* ChainManager::writeSolution()
{
	return "";
}

void ChainManager::writeMissingList(int maxReport)
{
	printf("S 0 0\nMIN_N = %d\t MAX_N = %d\tMISSING LIST: (%d total) \n--\t", this->minN, this->maxN, this->num_missing);
	int num_missing = 0;
	for ( int i = this->minN + 1; i < this->maxN; i += 2 )
	{
		if ( this->isNumberFound(i) == false )
		{
			printf("%10d ", i);
			num_missing++;

			if ( (num_missing % 10) == 0 )
			{
				printf("\n--\t");
			}

			if ( num_missing >= maxReport )
			{
				break;
			}
		}
	}
	printf("\n");
	fflush(stdout);
}

/**
 * writeStatistics -- create a buffer that contains a
 * 	description of the solution.
 *
 * Statistics take the following format in each line:
 *
 * T [TYPE] [ID] [VALUE]
 *
 * @return a string of data allocated with malloc().
 * 	It will be deleted using free().
 */
char* ChainManager::writeStatistics()
{
	/* Here, write the statistics for MISSING NUMBERS */
	if ( (this->deepeningMode & DEEPEN_GENERATE) == 0 )
	{
		this->writeMissingList(10000);
	}

	return this->writeBaseStats();
}

bool ChainManager::isNumberFound(int num)
{
	if ( num <= this->minN || num > this->maxN )
	{
		return false;
	}

	if ( (num % 2) == 0 )
	{
		return false;
	}

	int pos = num - this->minN;
	int odd = (pos - 1) / 2;

	int bit = odd % 6;
	int byte = odd / 6;

	return (((this->found_numbers[byte]) & (1 << bit)) != 0);
}

bool ChainManager::setNumberFound(int num)
{
	if ( num <= this->minN || num > this->maxN )
	{
		return false;
	}

	if ( (num % 2) == 0 )
	{
		return false;
	}

	if ( this->isNumberFound(num) )
	{
		return true;
	}

	this->num_missing = this->num_missing - 1;
	int pos = num - this->minN;
	int odd = (pos - 1) / 2;

	int bit = odd % 6;
	int byte = odd / 6;
	this->found_numbers[byte] = (this->found_numbers[byte]) | (1 << bit);

	return true;
}

